/*
TGA RGB to 4:2:2 YCbCr (UYVY) conversion
*/

#include <stdio.h>
#include <stdlib.h>

#define	TRUE		1
#define	FALSE		0

void conv444to422(unsigned char *src, unsigned char *dst, int width, int height);

unsigned char *Clip;

int main(int argc, char **argv)
{
	FILE	*fp;
	FILE	*fpout;
	int	length, i, j, horz, vert, index = 0;
	int	vertical_flip = TRUE;
	int	r, g, b;
	double	y, u, v;
	double	cr, cg, cb;
	double	cr1, cg1, cb1;
	double	cr2, cg2, cb2;
	unsigned char	*yp, *up, *vp;
	int	matrix_coefficients = 1;
	unsigned int	py, pcb, pcr;
	static unsigned char	buffer_422output[3840 * 2160 * 2];
	static unsigned char	buffer_422Cb[3840 * 2160 / 2];
	static unsigned char	buffer_422Cr[3840 * 2160 / 2];
	static unsigned char	buffer_444Y[3840 * 2160];
	static unsigned char	buffer_444Cb[3840 * 2160];
	static unsigned char	buffer_444Cr[3840 * 2160];
	static unsigned char	buffer_444input[3840 * 2160 * 3];
	static double coef[7][3] = {
		{0.2126, 0.7152, 0.0722},   /* ITU-R Rec. 709 */
		{0.299, 0.587, 0.114},      /* unspecified */
		{0.299, 0.587, 0.114},      /* reserved */
		{0.30,  0.59,  0.11},       /* FCC */
		{0.299, 0.587, 0.114},      /* ITU-R Rec. 624-4 System B, G */
		{0.299, 0.587, 0.114},      /* SMPTE 170M */
		{0.212, 0.701, 0.087}};     /* SMPTE 240M (1987) */
	static double coefcb[7][3] = {
		{-0.1172, -0.3942, 0.5114}, /* ITU-R Rec. 709 */
		{-0.1726, -0.3389, 0.511},  /* unspecified */
		{-0.1726, -0.3389, 0.511},  /* reserved */
		{-0.173, -0.339, 0.511},    /* FCC */
		{-0.1726, -0.3389, 0.511},  /* ITU-R Rec. 624-4 System B, G */
		{-0.1726, -0.3389, 0.511},  /* SMPTE 170M */
		{-0.119, -0.393, 0.511}};   /* SMPTE 240M (1987) */
	static double coefcr[7][3] = {
		{0.5114, -0.4646, -0.0468}, /* ITU-R Rec. 709 */
		{0.511, -0.4283, -0.0832},  /* unspecified */
		{0.511, -0.4283, -0.0832},  /* reserved */
		{0.511, -0.431, -0.081},    /* FCC */
		{0.511, -0.4283, -0.0832},  /* ITU-R Rec. 624-4 System B, G */
		{0.511, -0.4283, -0.0832},  /* SMPTE 170M */
		{0.511, -0.455, -0.056}};   /* SMPTE 240M (1987) */

	if (argc != 5) {
		fprintf(stderr, "usage: rgbtouyvy <horz> <vert> <infile> <outfile>\n");
		exit(-1);
	}

	/*--- open binary file (for parsing) ---*/
	fp = fopen(argv[3], "rb");
	if (fp == 0) {
		fprintf(stderr, "Cannot open input file <%s>\n", argv[3]);
		exit(-1);
	}

	/*--- open binary file (for parsing) ---*/
	fpout = fopen(argv[4], "wb");
	if (fpout == 0) {
		fprintf(stderr, "Cannot open output file <%s>\n", argv[4]);
		exit(-1);
	}

	/* Clip table */
	if (!(Clip=(unsigned char *)malloc(1024)))
		printf("Clip[] malloc failed\n");

	Clip += 384;

	for (i = -384; i < 640; i++)
		Clip[i] = (i<0) ? 0 : ((i>255) ? 255 : i);

	horz = atoi(argv[1]);
	vert = atoi(argv[2]);
	length = fread(&buffer_444input[0], 1, 18, fp);
	if (buffer_444input[17] & 0x20)
	{
		vertical_flip = FALSE;
	}
	length = fread(&buffer_444input[0], 1, (horz * vert * 3), fp);

	i = matrix_coefficients;
	cr = coef[i-1][0];
	cg = coef[i-1][1];
	cb = coef[i-1][2];
	cr1 = coefcb[i-1][0];
	cg1 = coefcb[i-1][1];
	cb1 = coefcb[i-1][2];
	cr2 = coefcr[i-1][0];
	cg2 = coefcr[i-1][1];
	cb2 = coefcr[i-1][2];

	if (vertical_flip == FALSE)
	{
		index = 0;
	}
	else
	{
		index = ((horz * (vert - 1)) * 3);
	}

	for (i = 0; i < vert; i++)
	{
		yp = &buffer_444Y[0] + i*horz;
		up = &buffer_444Cb[0] + i*horz;
		vp = &buffer_444Cr[0] + i*horz;

		for (j =0 ; j < horz; j++)
		{
			b = buffer_444input[index++];
			g = buffer_444input[index++];
			r = buffer_444input[index++];

			/* convert to YUV */
			y = cr * r + cg * g + cb * b;
			u = cr1 * r + cg1 * g + cb1 * b;
			v = cr2 * r + cg2 * g + cb2 * b;
			yp[j] = Clip[(int)(y + 0.5)];
			up[j] = Clip[(int)(u + 128.5)];
			vp[j] = Clip[(int)(v + 128.5)];
		}
		if (vertical_flip)
		{
			index = index - (horz * 6);
		}
	}
	conv444to422(&buffer_444Cb[0], &buffer_422Cb[0], horz, vert);
	conv444to422(&buffer_444Cr[0], &buffer_422Cr[0], horz, vert);
	py = 0;
	pcb = 0;
	pcr = 0;
	for (i = 0; i < (horz * vert * 2); i++)
	{
		if ((i % 2) == 1)
		{
			buffer_422output[i] = buffer_444Y[py++];
		}
		else if ((i % 4) == 0)
		{
			buffer_422output[i] = buffer_422Cb[pcb++];
		}
		else
		{
			buffer_422output[i] = buffer_422Cr[pcr++];
		}
	}
	fwrite(&buffer_422output[0], 1, (horz * vert * 2), fpout);
	fclose(fp);
	fclose(fpout);
	return 0;
}

/* horizontal filter and 2:1 subsampling */
void conv444to422(unsigned char *src, unsigned char *dst, int width, int height)
{
  int i, j, im5, im4, im3, im2, im1, ip1, ip2, ip3, ip4, ip5, ip6;

  if (0)
  {
    for (j=0; j<height; j++)
    {
      for (i=0; i<width; i+=2)
      {
        im5 = (i<5) ? 0 : i-5;
        im4 = (i<4) ? 0 : i-4;
        im3 = (i<3) ? 0 : i-3;
        im2 = (i<2) ? 0 : i-2;
        im1 = (i<1) ? 0 : i-1;
        ip1 = (i<width-1) ? i+1 : width-1;
        ip2 = (i<width-2) ? i+2 : width-1;
        ip3 = (i<width-3) ? i+3 : width-1;
        ip4 = (i<width-4) ? i+4 : width-1;
        ip5 = (i<width-5) ? i+5 : width-1;
        ip6 = (i<width-5) ? i+6 : width-1;

        /* FIR filter with 0.5 sample interval phase shift */
        dst[i>>1] = Clip[(int)(228*(src[i]+src[ip1])
                         +70*(src[im1]+src[ip2])
                         -37*(src[im2]+src[ip3])
                         -21*(src[im3]+src[ip4])
                         +11*(src[im4]+src[ip5])
                         + 5*(src[im5]+src[ip6])+256)>>9];
      }
      src+= width;
      dst+= width>>1;
    }
  }
  else
  {
    /* MPEG-2 */
    for (j=0; j<height; j++)
    {
      for (i=0; i<width; i+=2)
      {
        im5 = (i<5) ? 0 : i-5;
        im3 = (i<3) ? 0 : i-3;
        im1 = (i<1) ? 0 : i-1;
        ip1 = (i<width-1) ? i+1 : width-1;
        ip3 = (i<width-3) ? i+3 : width-1;
        ip5 = (i<width-5) ? i+5 : width-1;

        /* Original FIR filter coefficients (*512): 22 0 -52 0 159 256 159 0 -52 0 22 */
        /* It's a bug. Filter coefficients must add up to 512!!! */

        /* FIR filter coefficients (*512): 21 0 -52 0 159 256 159 0 -52 0 21 */
        dst[i>>1] = Clip[(int)(  21*(src[im5]+src[ip5])-52*(src[im3]+src[ip3])
                         +159*(src[im1]+src[ip1])+256*src[i]+256)>>9];
      }
      src+= width;
      dst+= width>>1;
    }
  }
}
